from rest_framework import pagination, serializers
from rest_framework.filters import OrderingFilter, SearchFilter

from django_filters.rest_framework import DjangoFilterBackend
from django.core.exceptions import FieldDoesNotExist

from django.db.models.fields.reverse_related import ManyToOneRel, OneToOneRel, ManyToManyRel

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.fields import NOT_PROVIDED


class NullToDefaultMixin(object):

    def __init__(self, *args, **kwargs):
        super(NullToDefaultMixin, self).__init__(*args, **kwargs)
        for field in self.Meta.fields:
            try:
                model_field = self.Meta.model._meta.get_field(field)
                if hasattr(model_field, 'default') and model_field.default != NOT_PROVIDED:
                    self.fields[field].allow_null = True
            except FieldDoesNotExist:
                pass

    def validate(self, data):
        for field in self.Meta.fields:
            try:
                model_field = self.Meta.model._meta.get_field(field)
                if hasattr(model_field, 'default') and model_field.default != NOT_PROVIDED and \
                        data.get(field, NOT_PROVIDED) is None:
                    data.pop(field)
            except FieldDoesNotExist:
                pass

        return super(NullToDefaultMixin, self).validate(data)


def serializer_factory(endpoint=None, fields=None, base_class=None, model=None):

    if model is not None:
        assert endpoint is None, "You cannot specify both a model and an endpoint"
        from .endpoints import Endpoint
        endpoint = Endpoint(model=model)
    else:
        assert endpoint is not None, "You have to specify either a model or an endpoint"

    if base_class is None:
        base_class = endpoint.base_serializer

    meta_attrs = {
        'model': endpoint.model,
        'fields': fields if fields is not None else endpoint.get_fields_for_serializer()
    }
    meta_parents = (object, )
    if hasattr(base_class, 'Meta'):
        meta_parents = (base_class.Meta, ) + meta_parents

    Meta = type('Meta', meta_parents, meta_attrs)

    cls_name = '{}Serializer'.format(endpoint.model.__name__)
    cls_attrs = {
        'Meta': Meta,
    }

    for meta_field in meta_attrs['fields']:
        if meta_field not in base_class._declared_fields:
            try:
                model_field = endpoint.model._meta.get_field(meta_field)
                if isinstance(model_field, OneToOneRel):
                    cls_attrs[meta_field] = serializers.PrimaryKeyRelatedField(read_only=True)
                elif isinstance(model_field, ManyToOneRel):
                    cls_attrs[meta_field] = serializers.PrimaryKeyRelatedField(many=True, read_only=True)
                elif isinstance(model_field, ManyToManyRel):
                    # related ManyToMany should not be required
                    cls_attrs[meta_field] = serializers.PrimaryKeyRelatedField(
                        many=True,
                        required=False,
                        queryset=model_field.related_model.objects.all()
                    )
            except FieldDoesNotExist:
                cls_attrs[meta_field] = serializers.ReadOnlyField()

    try:
        return type(cls_name, (NullToDefaultMixin, base_class, ), cls_attrs)
    except TypeError:
        # MRO issue, let's try the other way around
        return type(cls_name, (base_class, NullToDefaultMixin, ), cls_attrs)


def pagination_factory(endpoint):
    pg_cls_name = '{}Pagination'.format(endpoint.model.__name__)

    page_size = getattr(endpoint, 'page_size', None)
    pg_cls_attrs = {
        'page_size': page_size if page_size is not None else settings.REST_FRAMEWORK.get('PAGE_SIZE', 50),
    }

    if hasattr(endpoint, 'pagination_template'):
        pg_cls_attrs['template'] = endpoint.pagination_template

    BasePagination = getattr(endpoint, 'base_pagination_class', pagination.PageNumberPagination)
    if issubclass(BasePagination, pagination.PageNumberPagination):
        pg_cls_attrs['page_size_query_param'] = getattr(endpoint, 'page_size_query_param', 'page_size')
        for param in ('django_paginator_class', 'page_query_param', 'max_page_size', 'last_page_string',
                      'page_size'):
            if getattr(endpoint, param, None) is not None:
                pg_cls_attrs[param] = getattr(endpoint, param)
    elif issubclass(BasePagination, pagination.LimitOffsetPagination):
        pg_cls_attrs.pop('page_size')
        for param in ('default_limit', 'limit_query_param', 'offset_query_param', 'max_limit'):
            if getattr(endpoint, param, None) is not None:
                pg_cls_attrs[param] = getattr(endpoint, param)
    elif issubclass(BasePagination, pagination.CursorPagination):
        for param in ('page_size', 'cursor_query_param', 'ordering'):
            if getattr(endpoint, param, None) is not None:
                pg_cls_attrs[param] = getattr(endpoint, param)
    else:
        raise ImproperlyConfigured('base_pagination_class needs to be a subclass of one of the following:'
                                   'PageNumberPagination, LimitOffsetPagination, CursorPagination')

    return type(pg_cls_name, (BasePagination, ), pg_cls_attrs)


def filter_factory(endpoint):

    base_class = endpoint.base_filter_class

    cls_name = '{}FilterSet'.format(endpoint.model.__name__)

    meta_attrs = {
        'model': endpoint.model,
        'fields': [field if not isinstance(field, dict) else field.get('key', field['name'])
                   for field in endpoint.filter_fields]
    }

    meta_parents = (object, )
    if hasattr(base_class, 'Meta'):
        meta_parents = (base_class.Meta, ) + meta_parents

    Meta = type('Meta', meta_parents, meta_attrs)

    cls_attrs = {
        'Meta': Meta,
    }

    return type(cls_name, (base_class, ), cls_attrs)


def viewset_factory(endpoint):
    from .endpoints import BaseEndpoint

    base_viewset = endpoint.get_base_viewset()

    cls_name = '{}ViewSet'.format(endpoint.model.__name__)
    tmp_cls_attrs = {
        'serializer_class': endpoint.get_serializer(),
        'queryset': endpoint.model.objects.all() if getattr(endpoint, 'queryset', None) is None else endpoint.queryset,
        'endpoint': endpoint,
        '__doc__': base_viewset.__doc__
    }

    cls_attrs = {
        key: value
        for key, value in tmp_cls_attrs.items() if key == '__doc__' or
        getattr(base_viewset, key, None) is None
    }

    if 'filter_class' in cls_attrs or 'base_filter_class' in cls_attrs:
        cls_attrs.pop('filter_fields', None)

    if endpoint.permission_classes is not None:
        cls_attrs['permission_classes'] = endpoint.permission_classes

    filter_backends = getattr(endpoint.get_base_viewset(), 'filter_backends', ())
    if filter_backends is None:
        filter_backends = []
    else:
        filter_backends = list(filter_backends)

    for filter_type, backend in (
        ('filter_fields', DjangoFilterBackend),
        ('search_fields', SearchFilter),
        ('ordering_fields', OrderingFilter),
    ):
        if hasattr(endpoint, 'get_{}'.format(filter_type)):
            method = getattr(endpoint, 'get_{}'.format(filter_type))
            try:
                val = method(check_viewset_if_none=False)
            except TypeError:
                val = method(request=None, check_viewset_if_none=False)

        else:
            val = []
        if val is not None and val != []:
            filter_backends.append(backend)

            if filter_type == 'filter_fields':
                filter_values = [field['name'] if isinstance(field, dict) else field
                                 for field in val]
                # Pre django-filters 22.1
                cls_attrs['filter_fields'] = filter_values
                # django-filters 22.1+
                cls_attrs['filterset_fields'] = filter_values
            elif filter_type == 'ordering_fields':
                cls_attrs['ordering_fields'] = [field['filter'] if isinstance(field, dict) else field
                                                for field in val]
            else:
                cls_attrs[filter_type] = getattr(endpoint, filter_type)

    if hasattr(endpoint, 'filter_class'):
        cls_attrs['filter_class'] = endpoint.filter_class
    elif hasattr(endpoint, 'base_filter_class'):
        cls_attrs['filter_class'] = filter_factory(endpoint)

    if DjangoFilterBackend not in filter_backends and (hasattr(endpoint, 'filter_class') or
                                                       hasattr(base_viewset, 'filter_class') or
                                                       hasattr(endpoint, 'base_filter_class')):
        filter_backends.append(DjangoFilterBackend)

    if len(filter_backends) > 0:
        cls_attrs['filter_backends'] = filter_backends

    if hasattr(endpoint, 'pagination_class'):
        cls_attrs['pagination_class'] = endpoint.pagination_class
    else:
        cls_attrs['pagination_class'] = pagination_factory(endpoint)

    rv = type(cls_name, (endpoint.get_base_viewset(),), cls_attrs)

    black_list = dir(BaseEndpoint)
    for method_name in dir(endpoint):
        if method_name not in black_list:
            method = getattr(endpoint, method_name)
            if getattr(method, 'action_type', None) in ['custom', 'bulk', 'list']:
                setattr(rv, method_name, method)

    return rv



from django.conf import settings as django_settings
from django.utils.module_loading import import_string
from django.utils.text import capfirst

from inflector import Inflector

from .app_settings import settings
from .get_field_dict import get_field_dict  # NoQA


inflector_language = import_string(settings.INFLECTOR_LANGUAGE)
inflector = Inflector(inflector_language)


def action_kwargs(icon_class, btn_class, text, func, kwargs):

    kwargs.update({
        'icon_class': icon_class if icon_class is not None else settings.ACTION_ICON_CLASS,
        'btn_class': btn_class if btn_class is not None else settings.ACTION_BTN_CLASS,
    })

    if text is None:
        kwargs['text'] = capfirst(func.__name__.lower())
    else:
        kwargs['text'] = text

    return kwargs


def get_languages():
    if django_settings.USE_I18N:
        return [
            x[0] for x in getattr(django_settings, 'LANGUAGES', [[django_settings.LANGUAGE_CODE]])
        ]
    else:
        return None



from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string

from export_app import settings
from export_app.base import SerializerExporterWithFields, ModelNotFoundException


class Command(SerializerExporterWithFields, BaseCommand):
    help = 'Export DRF serializer definition to a frontend model definition'

    def add_arguments(self, parser):
        parser.add_argument('wizard_endpoint', nargs='*',
                            help="The shorthand url(s) of the endpoint(s) for which you'd like to export models."
                                 " Eg: 'sample/products'")
        parser.add_argument('--all', default=False, action='store_true',
                            help="Export all models corresponding to endpoints registered with your router")
        parser.add_argument('--adapter_name', default=settings.ADAPTER,
                            help='Defaults to {}'.format(settings.ADAPTER))
        parser.add_argument('--target_app', default=None,
                            help='Force all relationships to the target_app instead of the model\'s app')
        parser.add_argument('--router', default=None,
                            help='Defaults to {}'.format(settings.ROUTER_PATH))
        parser.add_argument('--noinput', default=False, action='store_true')

    def handle(self, *args, **options):
        target_app = options['target_app']
        adapter_name = options['adapter_name']
        if '.' not in adapter_name:
            adapter_name = 'export_app.adapters.{}'.format(adapter_name)
        Adapter = import_string(adapter_name)
        adapter = Adapter()

        if options['router'] is not None:
            self.router = import_string(options['router'])
        endpoints = options['wizard_endpoint']

        if len(endpoints) == 0 and not options['all']:
            self.print_help('drf_auto_endpoint', 'export')
            return
        elif options['all'] and len(endpoints) > 0:
            raise CommandError('You need to specify either wizard_endpoint(s) or use --all but you'
                               ' cannot use both at the same time')
        elif options['all']:
            endpoints = []
            for path, endpoint in getattr(self.router, '_endpoints', {}).items():
                viewset = endpoint.get_viewset()
                for method in dir(viewset):
                    if hasattr(getattr(viewset, method), 'wizard') and getattr(viewset, method).wizard:
                        endpoints.append('{}/{}'.format(path, method))

        for endpoint in endpoints:
            print('Exporting {} using {}'.format(endpoint, adapter_name))
            base_name, method_name = endpoint.rsplit('/', 1)

            try:
                viewset, model_name, application_name = self.get_viewset_for_basename(base_name)
                serializer = getattr(viewset, method_name).serializer
                serializer_instance = serializer()
            except ModelNotFoundException as e:
                raise CommandError('No viewset found for {}'.format(e.model))

            fields, rels = [], []

            if adapter.requires_fields:
                fields, rels = self.get_fields_for_model(None, serializer_instance, adapter,
                                                         target_app)

            belongsTo = False
            hasMany = False

            for rel in rels:
                if rel['type'] == 'belongsTo':
                    belongsTo = True
                    if hasMany:
                        break
                else:
                    hasMany = True
                    if belongsTo:
                        break

            base_name = 'wizard/{}'.format(base_name)
            context = {
                'endpoint': 'wizard/{}'.format(endpoint),
                'model_name': method_name,
                'application_name': base_name,
                'fields': fields,
                'rels': rels,
                'belongsTo': belongsTo,
                'hasMany': hasMany,
                'target_app': target_app,
                'api_base': settings.BACK_API_BASE,
                'pagination_container': None,
                'updir': 3,
            }

            print('writing {}/{}'.format(base_name, method_name))
            adapter.write_to_file(base_name, method_name, context, options['noinput'])



from django.core.exceptions import FieldDoesNotExist
from django.db.models.fields import NOT_PROVIDED
from django.utils.module_loading import import_string

from rest_framework import serializers, relations
from rest_framework.fields import empty

from inflector import Inflector

from .app_settings import settings


inflector_language = import_string(settings.INFLECTOR_LANGUAGE)
inflector = Inflector(inflector_language)


class GetFieldDict():

    def __call__(self, *args, **kwargs):
        return self.dict_for_field(*args, **kwargs)

    def get_validation_attrs(self, instance_field):
        rv = {}

        attrs_to_validation = {
            'min_length': 'min',
            'max_length': 'max',
            'min_value': 'min',
            'max_value': 'max'
        }
        for attr_name, validation_name in attrs_to_validation.items():
            if getattr(instance_field, attr_name, None) is not None:
                rv[validation_name] = getattr(instance_field, attr_name)

        return rv

    def get_read_only(self, name, field_instance):
        if name == '__str__':
            return True

        if field_instance.read_only and not isinstance(field_instance, serializers.ManyRelatedField):
            return True

        return False

    def get_base_dict_for_field(self, name, field_instance, serializer, translated_fields,
                                serializer_instance):

        read_only = self.get_read_only(name, field_instance)

        return {
            'key': name,
            'type': settings.WIDGET_MAPPING[field_instance.__class__.__name__],
            'read_only': read_only,
            'ui': {
                'label': name.title().replace('_', ' ')
                if name != '__str__'
                else serializer_instance.Meta.model.__name__,
            },
            'validation': {
                'required': field_instance.required,
            },
            'extra': {},
            'translated': name in translated_fields
        }

    def update_annotations(self, rv, name, field_instance, fields_annotation):

        if fields_annotation and name in fields_annotation:
            if 'placeholder' in fields_annotation[name]:
                rv['ui']['placeholder'] = fields_annotation[name]['placeholder']
            if 'help' in fields_annotation[name]:
                rv['ui']['help'] = fields_annotation[name]['help']
        if field_instance.help_text is not None and 'help' not in rv['ui']:
            rv['ui']['help'] = field_instance.help_text

    def get_model_field(self, field_instance, model):
        if model is None:
            return None
        try:
            return model._meta.get_field(field_instance.source)
        except FieldDoesNotExist:
            return None

    def update_default_from_model(self, rv, model_field):
        if model_field is None:
            return
        if hasattr(model_field, 'default') and model_field.default != NOT_PROVIDED:
            rv['default'] = model_field.default

    def update_default_from_serializer(self, rv, field_instance):
        if field_instance.default == empty:
            return
        rv['default'] = field_instance.default

    def normalize_default(self, rv, field_instance):
        if 'default' in rv:
            if callable(rv['default']):
                rv['default'] = rv['default']()

        # TODO: handle relationships

    def update_label(self, rv, model_field):
        if model_field is None:
            return

        try:
            rv['ui']['label'] = model_field.verbose_name
        except AttributeError:
            rv['ui']['label'] = model_field.name

    def update_extra(self, rv, field):
        if not isinstance(field, dict):
            return

        extra = rv['extra']
        extra.update(field.get('extra', {}))
        rv.update(field)
        rv['extra'] = extra

    def update_related_endpoint(self, rv, related_model):
        rv['related_endpoint'] = {
            'app': related_model._meta.app_label,
            'singular': related_model._meta.model_name.lower(),
            'plural': inflector.pluralize(related_model._meta.model_name.lower())
        }

    def set_choices_from_qs(self, rv, qs, key_attr='id'):
        rv['type'] = settings.WIDGET_MAPPING['choice']

        rv['choices'] = [
            {
                'label': record.__str__(),
                'value': getattr(record, key_attr)
            } for record in qs.all()
        ]

    def update_realtionship_from_model(self, rv, model_field, foreign_key_as_list):
        if model_field is None:
            return

        related_model = getattr(model_field, 'related_model', None)
        if related_model is None:
            return

        if model_field.__class__.__name__ == 'ManyToManyRel':
            rv['validation']['required'] = False

        if not foreign_key_as_list:
            rv['type'] = settings.WIDGET_MAPPING[model_field.__class__.__name__]
            self.update_related_endpoint(rv, related_model)
        else:
            # FIXME: we may not need this code as the serializer field has a 'choices' attribute
            qs = related_model.objects

            key_attr = 'pk'
            if hasattr(model_field, 'to_fields') and model_field.to_fields is not None \
                    and len(model_field.to_fields) > 0:
                key_attr = model_field.to_fields[0]

            self.set_choices_from_qs(rv, qs, key_attr)

    def update_relationship_from_serializer(self, rv, field_instance, foreign_key_as_list):
        if not isinstance(field_instance, (relations.PrimaryKeyRelatedField, relations.ManyRelatedField,
                                           relations.SlugRelatedField)):
            return

        if not hasattr(field_instance, 'queryset') or field_instance.queryset is None:
            return

        related_model = field_instance.queryset.model

        if not foreign_key_as_list:
            self.update_related_endpoint(rv, related_model)
        else:
            if not hasattr(field_instance, 'queryset') or field_instance.queryset is None:
                return
            # FIXME: we may not need this code as the serializer field has a 'choices' attribute
            self.set_choices_from_qs(rv, field_instance.queryset)

    def update_choices_from_serializer(self, rv, field_instance, force=False):
        if field_instance.__class__.__name__ != 'ChoiceField' and force is False:
            return

        if not hasattr(field_instance, 'choices'):
            return

        if rv.get('related_endpoint', None) is not None and force is False:
            return

        rv['type'] = settings.WIDGET_MAPPING['choice']
        rv['choices'] = [
            {
                'label': v,
                'value': k,
            } for k, v in field_instance.choices.items()
        ]

    def dict_for_field(self, field, serializer, translated_fields=None, fields_annotation=False,
                       model=None, foreign_key_as_list=False):
        if translated_fields is None:
            translated_fields = []

        serializer_instance = serializer()
        name = field['name'] if isinstance(field, dict) else field
        try:
            field_instance = serializer_instance.fields[name]
        except KeyError:
            return {'key': name}

        model_field = self.get_model_field(field_instance, model)

        rv = self.get_base_dict_for_field(name, field_instance, serializer, translated_fields, serializer_instance)
        self.update_annotations(rv, name, field_instance, fields_annotation)

        self.update_default_from_model(rv, model_field)
        self.update_default_from_serializer(rv, field_instance)
        self.normalize_default(rv, field_instance)

        self.update_label(rv, model_field)

        self.update_realtionship_from_model(rv, model_field, foreign_key_as_list)
        self.update_relationship_from_serializer(rv, field_instance, foreign_key_as_list)

        self.update_choices_from_serializer(rv, field_instance)

        rv['validation'].update(self.get_validation_attrs(field_instance))

        self.update_extra(rv, field)
        return rv


get_field_dict = GetFieldDict()



from django.urls import re_path
from export_app import settings
from export_app.views import EmberModelView, WizardModelView


urlpatterns = [
    re_path(r'^wizard/(?P<model>[\w\/_-]+)\.js', WizardModelView.as_view(), name='{}-wizard'.format(
        settings.URL_NAME
    )),
    re_path(r'^(?P<model>[\w\/_-]+)\.js', EmberModelView.as_view(), name=settings.URL_NAME),
]



from django.conf import settings as django_settings


DEFAULT_SETTINGS = {
    'ROUTER_PATH': 'urls.router',
    'URL_NAMESPACE': None,
    'URL_NAME': 'djember_model',
    'FRONT_APPLICATION_NAME': 'appName',
    'FRONT_APPLICATION_PATH': '../front',
    'BACK_API_BASE': '/api/v1',
    'URL_FORMAT': '{app}/{model}',
    'FIELD_TYPE_MAPPING': {},
    'DEFAULT_MAPPING': None,
    'ADAPTER': 'export_app.adapters.EmberAdapter',
    'EXCLUDE': {},
    'IGNORE_BASE': False
}


class Settings(object):

    def __getattr__(self, name):
        if name not in DEFAULT_SETTINGS:
            raise AttributeError(name)
        return getattr(django_settings, 'EXPORTER_{}'.format(name), DEFAULT_SETTINGS[name])


settings = Settings()



from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string

from export_app import settings
from export_app.base import SerializerExporterWithFields, ModelNotFoundException


class Command(SerializerExporterWithFields, BaseCommand):
    help = 'Export DRF serializer definition to a frontend model definition'

    def add_arguments(self, parser):
        parser.add_argument('model_endpoint', nargs='*',
                            help="The shorthand url(s) of the endpoint(s) for which you'd like to"
                                 " export models. Eg: 'sample/products'")
        parser.add_argument('--all', default=False, action='store_true',
                            help="Export all models corresponding to endpoints registered with your router")
        parser.add_argument('--adapter_name', default=settings.ADAPTER,
                            help='Defaults to {}'.format(settings.ADAPTER))
        parser.add_argument('--target_app', default=None,
                            help='Force all relationships to the target_app instead of the model\'s app')
        parser.add_argument('--router', default=None,
                            help='Defaults to {}'.format(settings.ROUTER_PATH))
        parser.add_argument('--noinput', default=False, action='store_true')

    def handle(self, *args, **options):
        target_app = options['target_app']
        adapter_name = options['adapter_name']
        if '.' not in adapter_name:
            adapter_name = 'export_app.adapters.{}'.format(adapter_name)
        Adapter = import_string(adapter_name)
        adapter = Adapter()

        excludes = settings.EXCLUDE
        if isinstance(excludes, dict):
            excludes = excludes.get(adapter_name.rsplit('.', 1)[-1], [])

        if options['router'] is not None:
            self.router = import_string(options['router'])
        endpoints = options['model_endpoint']

        if len(endpoints) == 0 and not options['all']:
            self.print_help('drf_auto_endpoint', 'export')
            return
        elif options['all'] and len(endpoints) > 0:
            raise CommandError('You need to specify either model_endpoint(s) or use --all but you '
                               'cannot use both at the same time')
        elif options['all']:
            endpoints = getattr(self.router, '_endpoints', {}).keys()

        for endpoint in endpoints:
            if endpoint not in excludes:
                endpoint_instance = self.get_endpoint_for_basename(endpoint)
                if adapter_name == settings.ADAPTER and \
                        getattr(endpoint_instance, 'default_export_adapter', None) is not None:
                    local_adapter_name = endpoint_instance.default_export_adapter.__name__
                    local_adapter = endpoint_instance.default_export_adapter()
                else:
                    local_adapter_name = adapter_name
                    local_adapter = adapter
                print('Exporting {} using {}'.format(endpoint, local_adapter_name))
                try:
                    if local_adapter.works_with in ['serializer', 'both']:
                        model, serializer_instance, model_name, application_name = \
                            self.get_serializer_for_basename(endpoint)

                        class BogusViewSet(object):
                            pagination_class = None

                        viewset = BogusViewSet()
                    if local_adapter.works_with in ['viewset', 'both']:
                        viewset, model_name, application_name = \
                            self.get_viewset_for_basename(endpoint, dasherize=local_adapter.dasherize)
                        viewset = viewset()
                except ModelNotFoundException as e:
                    raise CommandError('No viewset found for {}'.format(e.model))

                fields, rels = [], []

                if local_adapter.requires_fields:
                    fields, rels = self.get_fields_for_model(model, serializer_instance, local_adapter,
                                                             target_app, endpoint=endpoint_instance)

                belongsTo = False
                hasMany = False

                for rel in rels:
                    if rel['type'] == 'belongsTo':
                        belongsTo = True
                        if hasMany:
                            break
                    else:
                        hasMany = True
                        if belongsTo:
                            break

                if local_adapter.works_with in ['serializer', 'both']:
                    base = None
                    if not settings.IGNORE_BASE and model and \
                            model.__bases__[0].__name__ != 'django.db.models.base' and \
                            model.__bases__[0].__name__ != 'Model' and len(model.__bases__) == 1 and \
                            not model.__bases__[0]._meta.abstract:
                        base = (
                            model.__bases__[0].__name__.lower(),
                            model.__bases__[0].__module__.split('.')[0].lower().replace('_', '-')
                        )
                    context = {
                        'endpoint': endpoint,
                        'model_name': model_name,
                        'application_name': application_name,
                        'fields': fields,
                        'rels': rels,
                        'belongsTo': belongsTo,
                        'hasMany': hasMany,
                        'target_app': target_app,
                        'api_base': settings.BACK_API_BASE,
                        'base': base,
                    }

                    local_adapter.write_to_file(application_name, model_name, context, options['noinput'])
                else:
                    local_adapter.write_to_file(application_name, model_name, viewset, options['noinput'])

        adapter.rebuild_index()



import logging
import os

from django.conf import settings as django_settings
from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.utils.module_loading import import_string

from export_app import settings


class Command(BaseCommand):

    help = '''
Re-export already exported ember models
    '''

    logger = logging.getLogger('intranet')

    def handle(self, *args, **options):
        router = import_string(settings.ROUTER_PATH)
        endpoints = getattr(router, '_endpoints', {}).items()

        for endpoint, endpoint_instance in endpoints:
            model_name = endpoint_instance.singular_model_name
            application_name = endpoint_instance.application_name

            target = os.path.join(
                django_settings.BASE_DIR,
                settings.FRONT_APPLICATION_PATH,
                'app',
                'models',
                application_name.replace('_', '-'),
                '{}.js'.format(model_name).replace('_', '-')
            )

            if os.path.exists(target):
                call_command('export', endpoint, noinput=True)



import os
from collections import defaultdict

from django.conf import settings as django_settings
from django.template.loader import render_to_string
from django.utils.module_loading import import_string

from export_app import settings


try:
    input = raw_input
except NameError:
    pass


# from http://stackoverflow.com/questions/3203286/how-to-create-a-read-only-class-property-in-python#3203659
class classproperty(object):

    def __init__(self, getter):
        self.getter = getter

    def __get__(self, instance, owner):
        return self.getter(owner)


class BaseAdapter(object):

    FIELD_TYPE_MAPPING = {}
    DEFAULT_MAPPING = None

    requires_fields = False
    works_with = 'serializer'
    dasherize = False

    def __init__(self):
        assert not self.requires_fields or self.works_with != 'viewset', """
            Field information is provided by a serializer. Your adapter requires field information
            (requires_fields = True) and therefore also needs to "works_with" serializer.
            If you also require information from the viewset, please update "works_with" to "both"
            otherwise use "works_with = 'serializer'"
        """

    @classproperty
    def field_type_mapping(cls):
        default = cls.FIELD_TYPE_MAPPING
        default.update(settings.FIELD_TYPE_MAPPING)
        rv = defaultdict(lambda: cls.default_mapping)
        for k, v in default.items():
            rv[k] = v
        return rv

    @classproperty
    def default_mapping(cls):
        rv = settings.DEFAULT_MAPPING
        if rv is None:
            return cls.DEFAULT_MAPPING
        return rv

    def write_to_file(self, application_name, model_name, context, force_overwrite=False):
        raise NotImplementedError("You need to implement your Adapter")

    def create_dirs(self, *args):
        for directory in args:
            if not os.path.exists(directory):
                try:
                    os.makedirs(directory)
                except FileExistsError:
                    # someone created the directory in the meantime
                    # tis can happen in parrallel tests
                    pass

    def write_file(self, context, target_dir, filename, template, overwrite='confirm'):
        target_file = os.path.join(target_dir, filename)
        if os.path.exists(target_file):
            if not overwrite:
                return
            if overwrite == 'confirm':
                answer = 'anything else'
                while answer.lower() not in ('', 'yes', 'y', 'no', 'n'):
                    answer = input('{} already exists, do you want to overwrite it? [y/N] '.format(
                        target_file
                    ))
                    if answer.lower() in ('', 'no', 'n'):
                        return
        with open(target_file, 'w', encoding='utf-8') as f:
            output = render_to_string(template, context)
            f.write(output)

    def write_files(self, context, files):
        self.create_dirs(*[file[0] for file in files])
        for file in files:
            self.write_file(context, *file)

    def rebuild_index(self):
        pass

    @classmethod
    def adapt_fields_for_model(cls, fields, relationships):
        return fields, relationships


class EmberAdapter(BaseAdapter):

    FIELD_TYPE_MAPPING = {
        'BooleanField': 'boolean',
        'NullBooleanField': None,
        'IntegerField': 'number',
        'FloatField': 'number',
        'DecimalField': 'number',
        'ListField': None,
        'DictField': None,
        'JSONField': None,
        'PrimaryKeyRelatedField': 'belongsTo',
        'SlugRelatedField': 'belongsTo',
        'ManyRelatedField': 'hasMany',
        'DateField': 'nullable',
        'DateTimeField': 'nullable',
        'TimeField': 'nullable',
        'DurationField': 'duration',
        'UUIDField': 'nullable',
    }
    DEFAULT_MAPPING = 'string'
    requires_fields = True
    dasherize = True

    base_template_name = 'export_app/ember_model_base.js'
    template_name = 'export_app/ember_model.js'
    test_template_name = 'export_app/ember_model_test.js'
    dynamic_template_name = 'export_app/dynamic_model.js'

    def write_to_file(self, application_name, model_name, context, force_overwrite=False):
        context['application_name'] = context['application_name'].replace('_', '-')
        context['updir'] = context.get('updir', 1)
        base_target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                       'app', 'models', 'base', application_name.replace('_', '-'))
        target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                  'app', 'models', application_name.replace('_', '-'))
        test_target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                       'tests', 'unit', 'models', application_name.replace('_', '-'))

        filename = '{}.js'.format(model_name).replace('_', '-')
        test_filename = '{}-test.js'.format(model_name).replace('_', '-')

        files = [
            (base_target_dir, filename, self.base_template_name, True),
            (target_dir, filename, self.template_name, False),
            (test_target_dir, test_filename, self.test_template_name, 'confirm' if not force_overwrite else False)
        ]
        self.write_files(context, files)


class BaseMetadataAdapter(BaseAdapter):

    works_with = 'viewset'

    def render(self, data):
        from rest_framework.renderers import JSONRenderer

        renderer = JSONRenderer()
        return renderer.render(data)

    def get_metadata_from_viewset(self, viewset):

        if isinstance(viewset, dict):
            output = viewset
        else:
            MetadataClass = import_string(django_settings.REST_FRAMEWORK['DEFAULT_METADATA_CLASS'])
            output = MetadataClass().determine_metadata(None, viewset)

        return output

    def get_json(self, viewset):

        return self.render(self.get_metadata_from_viewset(viewset))


class MetadataAdapter(BaseMetadataAdapter):

    def write_to_file(self, application_name, model_name, viewset, force_overwrite=False):

        target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                  'data')
        if not os.path.exists(target_dir):
            os.makedirs(target_dir)

        filename = '{}-{}.json'.format(application_name, model_name)

        with open(os.path.join(target_dir, filename), 'wb') as f:
            f.write(self.get_json(viewset))


class MetadataES6Adapter(BaseMetadataAdapter):

    template_name = 'export_app/ember_metadata.js'
    index_template_name = 'export_app/ember_metadata_index.js'

    def walk_dir(self, base, ignore_index=False, prefix=''):
        from inflector import Inflector
        from django.utils.module_loading import import_string
        from drf_auto_endpoint.app_settings import settings as auto_settings

        inflector_language = import_string(auto_settings.INFLECTOR_LANGUAGE)
        inflector = Inflector(inflector_language)

        imports = []

        for item in os.listdir(base):
            filename = os.path.join(base, item)
            if os.path.isdir(filename):
                imports += self.walk_dir(filename, prefix=os.path.join(prefix, item.replace('_', '-')))
            elif item == 'index.js' and ignore_index:
                continue
            else:
                try:
                    base_name, extension = item.rsplit('.', 1)
                except ValueError:
                    # a file without extension
                    continue
                if extension != 'js':
                    continue
                imports.append((
                    os.path.join(prefix, inflector.pluralize(base_name)),
                    os.path.join(prefix, base_name).replace('/', '_').replace('-', '_'),
                    os.path.join('.', prefix.replace('-', '_'), base_name)
                ))
        return imports

    def rebuild_index(self):
        MetadataClass = import_string(django_settings.REST_FRAMEWORK['DEFAULT_METADATA_CLASS'])
        context = {}
        directory = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                 'app', 'data')
        context['items'] = self.walk_dir(directory, True)
        context['root_metadata'] = self.render(
            MetadataClass().determine_metadata(None, 'APIRootView')
        ).decode('utf-8')
        self.write_file(context, directory, 'index.js', self.index_template_name, True)

    def write_to_file(self, application_name, model_name, viewset, force_overwrite=False):
        target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                  'app', 'data', application_name)

        self.create_dirs(target_dir)
        filename = '{}.js'.format(model_name)

        output = self.get_json(viewset)

        context = {
            'json': output.decode('utf-8')
        }

        self.write_file(context, target_dir, filename, self.template_name, True)


class MobxAxiosAdapter(BaseAdapter):

    works_with = 'both'
    requires_fields = True

    config_template_name = 'export_app/mobxaxios_config.js'
    base_model_template_name = 'export_app/mobxaxios_base_model.js'
    model_base_template_name = 'export_app/mobxaxios_model_base.js'
    model_template_name = 'export_app/mobxaxios_model.js'
    base_store_template_name = 'export_app/mobxaxios_base_store.js'
    store_template_name = 'export_app/mobxaxios_store.js'

    def write_to_file(self, application_name, model_name, context, force_overwrite=False):
        base_target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH)
        config_target_dir = os.path.join(base_target_dir, 'config')
        model_target_dir = os.path.join(base_target_dir, 'models')
        model_base_target_dir = os.path.join(model_target_dir, 'base')
        store_target_dir = os.path.join(base_target_dir, 'stores')

        filename = '{}{}.js'.format(application_name, model_name)

        files = [
            (config_target_dir, 'axios-config.js', self.config_template_name, False),
            (store_target_dir, '_base.js', self.base_store_template_name, True),
            (store_target_dir, filename, self.store_template_name, 'confirm' if not force_overwrite else False),
            (model_base_target_dir, '_base.js', self.base_model_template_name, True),
            (model_base_target_dir, filename, self.model_base_template_name, True),
            (model_target_dir, filename, self.model_template_name, False)
        ]
        self.write_files(context, files)


class Angular2Adapter(BaseAdapter):

    FIELD_TYPE_MAPPING = {
        'BooleanField': 'boolean',
        'NullBooleanField': 'boolean',
        'IntegerField': 'number',
        'FloatField': 'number',
        'DecimalField': 'number',
        'ListField': 'any',
        'DictField': 'any',
        'JSONField': 'any',
        'PrimaryKeyRelatedField': 'belongsTo',
        'ManyRelatedField': 'hasMany',
    }
    DEFAULT_MAPPING = 'string'
    requires_fields = True

    base_template_name = 'export_app/angular2_model_base.ts'
    template_name = 'export_app/angular2_model.ts'
    service_base_template_name = 'export_app/angular2_service_base.ts'
    service_template_name = 'export_app/angular2_service.ts'

    def write_to_file(self, application_name, model_name, context, force_overwrite=False):
        file_model_name = model_name
        context['application_name'] = context['application_name']
        context['updir'] = context.get('updir', 1)
        target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH,
                                  'app', 'models', application_name)

        base_filename = '{}.base.ts'.format(file_model_name)
        filename = '{}.ts'.format(file_model_name)
        service_base_filename = '{}.service.base.ts'.format(file_model_name)
        service_filename = '{}.service.ts'.format(file_model_name)

        files = [
            (target_dir, base_filename, self.base_template_name, True),
            (target_dir, service_base_filename, self.service_base_template_name, True),
            (target_dir, filename, self.template_name, False if force_overwrite else 'confirm'),
            (target_dir, service_filename, self.service_template_name, False if force_overwrite else 'confirm'),
        ]
        self.write_files(context, files)


class VuexORMAxiosAdapter(BaseAdapter):

    FIELD_TYPE_MAPPING = {
        'ListField': None,
        'DictField': None,
        'JSONField': None,
        'BooleanField': 'boolean',
        'NullBooleanField': 'boolean',
        'IntegerField': 'number',
        'DecimalField': 'number',
        'PrimaryKeyRelatedField': 'number',
        'SludRelatedField': 'string',
        'UUIDField': 'uid',
        'ManyRelatedField': None,
    }
    DEFAULT_MAPPING = 'string'

    works_with = 'both'
    requires_fields = True

    database_template_name = 'export_app/vuexorm_axios_database.js'
    base_model_template_name = 'export_app/vuexorm_axios_base_model.js'
    model_base_template_name = 'export_app/vuexorm_axios_model_base.js'
    model_template_name = 'export_app/vuexorm_axios_model.js'

    def write_to_file(self, application_name, model_name, context, force_overwrite=False):
        base_target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH)
        base_model_target_dir = os.path.join(base_target_dir, 'models')
        model_target_dir = os.path.join(base_model_target_dir, application_name)

        filename = '{}.js'.format(model_name)

        files = [
            # (database_target_dir, 'database.js', self.database_template_name, True),
            (base_model_target_dir, 'Base.js', self.base_model_template_name, True),
            (os.path.join(model_target_dir, 'base'), filename, self.model_base_template_name, True),
            (model_target_dir, filename, self.model_template_name, False),
        ]

        self.write_files(context, files)

    def walk_dir(self, base, ignore_index=False, prefix=''):
        imports = []

        for item in os.listdir(base):
            filename = os.path.join(base, item)
            if os.path.isdir(filename):
                if item == 'base':
                    continue
                imports += self.walk_dir(filename, prefix=os.path.join(prefix, item.replace('_', '-')))
            elif item == 'Base.js':
                continue
            else:
                try:
                    base_name, extension = item.rsplit('.', 1)
                except ValueError:
                    # a file without extension
                    continue
                if extension != 'js':
                    continue
                imports.append((
                    os.path.join(prefix, base_name).replace('/', '_').replace('-', '_'),
                    os.path.join(prefix.replace('-', '_'), base_name)
                ))
        return imports

    def rebuild_index(self):
        base_target_dir = os.path.join(django_settings.BASE_DIR, settings.FRONT_APPLICATION_PATH)
        directory = os.path.join(base_target_dir, 'models')
        database_target_dir = os.path.join(base_target_dir, 'store')

        context = {
            'items': self.walk_dir(directory, True)
        }

        self.write_file(context, database_target_dir, 'database.js', self.database_template_name, True)
